#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from flask import abort, Blueprint
from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
    get_jwt_claims,
    get_jwt_identity,
    get_raw_jwt,
    jwt_required,
    jwt_refresh_token_required
)

from ... import blueprint_name
from ...extension.form import UserForm
from ...extension.model import Token, User

app = Blueprint(
    (
        app_name := blueprint_name(__file__)
    ),
    __name__,
    url_prefix=f'/{ app_name }/'
)


@app.route('/')
@jwt_required
def get():
    return get_jwt_claims()


@app.route('/', methods=['POST'])
def post():
    if (
        form := UserForm()
    ).validate_on_submit():
        error_parameter = {}

        if (
            user := User.read(
                User.username == (
                    username := form.username.data
                )
            )
        ):
            user = user[0]

            if not user.is_password(
                (password := form.password).data
            ):
                error_parameter[password.name] = (
                    f'{ password.name.title() } salah!'
                )
        else:
            error_parameter[form.username.name] = (
                f'{ username } tidak terdaftar!'
            )
    else:
        error_parameter = {
            key: value[0] if len(value) == 1 else value
            for key, value in form.errors.items()
        }

    if error_parameter:
        # Which HTTP code is best suited for validation errors: 400 or 422?
        # https://www.quora.com/Which-HTTP-code-is-best-suited-for-validation-errors-400-or-422
        abort(422, {'parameter': error_parameter})

    data = {
        key: func(user.id)
        for key, func in {
            'access_token': create_access_token,
            'refresh_token': create_refresh_token
        }.items()
    }
    Token.create(
        data['refresh_token']
    )
    return data, 201


@app.route('/', methods=['PUT'])
@jwt_refresh_token_required
def put():
    return {
        'access_token': create_access_token(
            get_jwt_identity()
        )
    }, 201


@app.route('/', methods=['DELETE'])
@jwt_refresh_token_required
def delete():
    Token.update(
        Token.jti == get_raw_jwt()['jti'],
        {'is_revoked': True}
    )
    return (
        f'{ User.read(get_jwt_identity()).username } berhasil logged out!',
        202
    )



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from functools import wraps

from flask import abort
from flask_jwt_extended import (
    get_jwt_identity,
    JWTManager,
    verify_jwt_in_request
)

from .model import Token, User

jwt = JWTManager()


def admin_required(fn):
    # Custom Decorators
    # https://flask-jwt-extended.readthedocs.io/en/stable/custom_decorators/
    @wraps(fn)
    def wrapper(*args, **kwargs):
        verify_jwt_in_request()

        if (
            user := User.read(
                get_jwt_identity()
            )
        ):
            result = fn(*args, **kwargs)

            if not user.is_admin:
                abort(403)
        else:
            result = revoked_token_loader()

        return result

    return wrapper


@jwt.user_claims_loader
def add_claims_to_access_token(identity):
    # Storing Data in Access Tokens
    # https://flask-jwt-extended.readthedocs.io/en/stable/add_custom_data_claims/
    (
        data := dict(
            User.read(identity)
        )
    ).pop('password', None)
    return data


@jwt.token_in_blacklist_loader
def check_if_token_in_blacklist(decrypted_token):
    # Blacklist and Token Revoking
    # https://flask-jwt-extended.readthedocs.io/en/stable/blacklist_and_token_revoking/
    return (
        # Memastikan token terdaftar
        len(
            Token.read(
                Token.jti == (
                    jti := decrypted_token['jti']
                )
            )
        ) == 0
        # Memastikan token tidak masuk daftar blacklist
        or len(
            Token.read(
                (Token.is_revoked == True) # noqa
                & (Token.jti == jti)
            )
        ) > 0
    )


# Changing Default Behaviors
# https://flask-jwt-extended.readthedocs.io/en/stable/changing_default_behavior/


def default_error(code=401, message=None):
    if code == 401:
        name = 'Unauthorized'
    elif code == 422:
        name = 'Unprocessable Entity'
    else:
        code = 500
        name = 'Internal Server Error'

    return {
        'error_message': message,
        'error_name': name
    }, code


@jwt.expired_token_loader
def expired_token_loader():
    return default_error(401, 'Token has expired.')


@jwt.invalid_token_loader
def invalid_token_loader(callback):
    return default_error(422, f'{ callback }.')


@jwt.revoked_token_loader
def revoked_token_loader():
    return default_error(401, 'Token has been revoked.')


@jwt.unauthorized_loader
def unauthorized_loader(callback):
    return default_error(401, f'{ callback }.')



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from os.path import abspath, isfile, join

from flask import (
    Blueprint,
    current_app,
    send_from_directory
)

app = Blueprint(
    'voxrow_route',
    __name__,
)


@app.route('/favicon.ico')
def favicon():
    return send_from_directory(
        join('static', 'img'),
        'logo.svg',
        mimetype='image/svg+xml'
    )


@app.route('/LICENSE')
def license():
    if not isfile(
        join(license_dir := current_app.root_path, 'LICENSE')
    ):
        license_dir = abspath(
            join(app.root_path, '..', '..')
        )

    return send_from_directory(
        license_dir,
        'LICENSE',
        mimetype='text/plain'
    )



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from flask_wtf import FlaskForm as _FlaskForm
from wtforms import BooleanField as _BooleanField


class FlaskForm(_FlaskForm):
    # CSRF Protection
    # https://wtforms.readthedocs.io/en/stable/csrf.html
    class Meta:
        csrf = False


class BooleanField(_BooleanField):
    # WTForms BooleanField Value And Validation
    # https://code.luasoftware.com/tutorials/python/wtforms-booleanfield-value-and-validation/

    def __init__(
        self,
        label=None,
        validators=None,
        false_values=(False, 'false', 'False', 'FALSE', 0, '0'),
        **kwargs
    ):
        # don't accept blank as False, so that default will trigger
        super().__init__(label, validators, false_values, **kwargs)

    def process_formdata(self, valuelist):
        if not valuelist or valuelist[0] == '' or valuelist[0] is None:
            self.data = self.default
        elif valuelist[0] in self.false_values:
            self.data = False
        else:
            self.data = True



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

# Packing and deploying a project to Pip
# https://medium.com/@ssbothwell/packing-and-deploying-a-project-to-pip-bcd628d02f6f
# Including files in source distributions with MANIFEST.in
# https://packaging.python.org/guides/using-manifest-in/

from setuptools import setup

setup(
    name='voxrow',
    packages=[
        'voxrow',
        'voxrow.modify',
        'voxrow.extension.route',
        'voxrow.extension',
        'voxrow.modify.flask_sqlalchemy'
    ],
    package_dir={
        'voxrow': '',
        'voxrow.extension': 'extension',
        'voxrow.extension.route': 'extension/route',
        'voxrow.modify': 'modify',
        'voxrow.modify.flask_sqlalchemy': 'modify/flask_sqlalchemy'
    },
    package_data={
        'voxrow': [
            'server/Makefile',
            'server/docker-compose/letsencrypt.yml',
            'server/install/*',
            'server/nginx/*'
        ]
    },
    version='0.9.18',
    description="VOXROW's library",
    author='Pipin Fitriadi',
    author_email='pipinfitriadi@gmail.com',
    license='MS-RSL',
    url='https://gitlab.com/voxrow/voxrow',
    install_requires=[
        'flask',
        'flask-jwt-extended',
        'Flask-SQLAlchemy',
        'Flask-WTF',
        'cryptography>=3.2.1',
        'sshtunnel'
    ],
    extras_require={
        'postgres': [
            # "psycopg2; sys_platform != 'darwin' or "
            # "platform_machine != 'arm64'",
            # "psycopg2-binary; sys_platform == 'darwin' "
            # "and platform_machine == 'arm64'"
            'psycopg2-binary'
        ]
    },
    download_url=(
        'https://gitlab.com/voxrow/voxrow/-/archive/master/'
        'voxrow-master.tar.gz'
    ),
    keywords=['voxrow', 'library', 'raw sql query'],
    python_requires='>=3.8.0',
    platforms='any'
)



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from datetime import datetime

from flask_jwt_extended import decode_token
from sqlalchemy import (
    Column,
    ForeignKey
)
from sqlalchemy.orm import validates
from sqlalchemy.schema import CheckConstraint
from sqlalchemy.types import (
    Boolean,
    DateTime,
    Integer,
    String
)

from ..modify.flask_sqlalchemy import DB_SCHEMA
from ..modify.flask_sqlalchemy import SQLAlchemy
from ..modify.flask_sqlalchemy.types import Password, Username

db = SQLAlchemy()
relationship = db.relationship
Model = db.Model


class User(Model):
    # Define minimum length for PostgreSQL string column with SQLAlchemy
    # https://stackoverflow.com/questions/50174325/define-minimum-length-for-postgresql-string-column-with-sqlalchemy
    __table_args__ = (
        CheckConstraint(
            'char_length(username) >= 5',
            name='username_min_length'
        ),
        CheckConstraint(
            'char_length(password) >= 8',
            name='password_min_length'
        ),
        # Table Configuration
        # https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/table_config.html
        {'schema': DB_SCHEMA}
    )

    id = Column(Integer, primary_key=True)
    is_admin = Column(
        Boolean,
        default=False,
        nullable=False
    )
    username = Column(
        # What is the maximum length of a facebook name?
        # https://stackoverflow.com/questions/8078939/what-is-the-maximum-length-of-a-facebook-name
        Username(50),
        unique=True,
        nullable=False
    )
    # What's the maximum number of characters permitted for the password of a
    # Google Account?
    # https://webapps.stackexchange.com/questions/57947/whats-the-maximum-number-of-characters-permitted-for-the-password-of-a-google-a
    password = Column(Password(100), nullable=False)
    # One-to-Many Relationships¶
    # https://flask-sqlalchemy.palletsprojects.com/en/2.x/models/
    token = relationship(
        'Token',
        backref='user',
        # {DetachedInstanceError} Parent instance <Car> is not bound to a
        # session; lazy load operation of attribute 'owner' cannot proceed
        # https://stackoverflow.com/questions/27701573/detachedinstanceerror-parent-instance-car-is-not-bound-to-a-session-lazy-lo
        lazy='subquery'
    )

    @validates('username')
    def validate_username(self, key, username) -> str:
        if len(username) < 5:
            raise ValueError('username too short')

        return username

    @validates('password')
    def validate_password(self, key, password) -> str:
        if len(password) < 8:
            raise ValueError('password too short')

        return password

    def is_password(self, password):
        return Password.is_correct(self.password, password)

    @classmethod
    def delete(cls, ident=None):
        users = cls.read(ident)

        def delete_token(user):
            for token in user.token:
                Token.delete(token.id)

        if ident and users:
            delete_token(users)
        else:
            for user in users:
                delete_token(user)

        return super().delete(ident)


class Token(Model):
    __table_args__ = {'schema': DB_SCHEMA}

    # Blacklist with a database
    # https://github.com/vimalloc/flask-jwt-extended/tree/master/examples/database_blacklist
    id = Column(Integer, primary_key=True)
    user_id = Column(
        Integer,
        # specifying schema name in model gives error
        # https://github.com/pallets/flask-sqlalchemy/issues/172
        ForeignKey(f'{ DB_SCHEMA }.user.id'),
        nullable=False
    )
    jti = Column(String(36), nullable=False)
    token_type = Column(String(10), nullable=False)
    expires = Column(DateTime, nullable=True)
    is_revoked = Column(
        Boolean,
        default=False,
        nullable=False
    )

    @classmethod
    def create(cls, token_or_row):
        if isinstance(token_or_row, str):
            token_or_row = cls(
                user_id=(
                    token := decode_token(token_or_row)
                )['identity'],
                jti=token['jti'],
                token_type=token['type'],
                expires=datetime.fromtimestamp(
                    token.get('exp')
                ) if token.get('exp') else None
            )

        cls.update(
            # flake8 complains on boolean comparison “==” in filter clause
            # https://stackoverflow.com/questions/18998010/flake8-complains-on-boolean-comparison-in-filter-clause
            (cls.is_revoked == False) # noqa
            & (
                (cls.user_id == token_or_row.user_id)
                | (cls.expires <= datetime.now())
            ),
            {'is_revoked': True}
        )
        super().create(token_or_row)



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from flask import abort, Blueprint
from sqlalchemy.exc import IntegrityError

from ... import blueprint_name
from ...extension.form import UserForm
from ...extension.model import User
from ...extension.token import admin_required

app = Blueprint(
    (
        app_name := blueprint_name(__file__)
    ),
    __name__,
    url_prefix=f'/{ app_name }/'
)


def to_dict(user):
    return {
        key: value
        for key, value in user
        if key != 'password'
    }


@app.route('/')
@app.route(route_id := '/<int(min=1):id>/')
@admin_required
def get(id=None):
    if id and (
        data := User.read(id)
    ):
        data = to_dict(data)
    elif not id:
        data = list(
            to_dict(user)
            for user in User.read()
        )
    else:
        abort(404)

    return data


@app.route('/', methods=['POST'])
@admin_required
def post():
    if (
        form := UserForm()
    ).validate_on_submit():
        error_parameter = {}
        is_admin = form.is_admin.data

        try:
            User.create(
                User(
                    username=(
                        username := form.username.data
                    ),
                    password=form.password.data,
                    is_admin=(
                        is_admin if is_admin else False
                    )
                )
            )
        except IntegrityError:
            error_parameter[form.username.name] = (
                f'{ username } sudah terdaftar!'
            )
    else:
        error_parameter = {
            key: value[0] if len(value) == 1 else value
            for key, value in form.errors.items()
        }

    if error_parameter:
        abort(422, {'parameter': error_parameter})

    return to_dict(
        User.read(User.username == username)[0]
    ), 201


@app.route(route_id, methods=['PUT'])
@admin_required
def put(id):
    if not User.read(id):
        abort(404)

    (
        form := UserForm()
    ).validate_on_submit()
    error_parameter = {}
    parameter = {}

    for key, value in form.data.items():
        if (
            value or (
                key == 'is_admin'
                and value is False
            )
        ):
            if (
                error := form.errors.get(key)
            ):
                error_parameter[key] = (
                    error[0]
                    if len(error) == 1
                    else error
                )
            else:
                parameter[key] = value

    try:
        if not error_parameter and parameter:
            User.update(User.id == id, parameter)
    except IntegrityError:
        error_parameter[form.username.name] = (
            f'{ form.username.data } sudah terdaftar!'
        )

    if error_parameter:
        abort(422, {'parameter': error_parameter})

    return to_dict(
        User.read(id)
    ), 201


@app.route(route_id, methods=['DELETE'])
@admin_required
def delete(id):
    if (
        data := User.read(id)
    ):
        User.delete(id)
    else:
        abort(404)

    return to_dict(data), 202
